#!/usr/bin/perl
# MML Parser and Decorder
# Programmed by Sasaji
# Version 0.2

use strict;

my $addr_start = 0xb800;

my $addr = $addr_start;

my $verb = 1;

my $outfile;
my %info_datas;
my @voice_datas;
my @score_datas;

set_datas();

my $fout;
if (!open($fout, "> ",$outfile)) {
	print "Cannot write to file.\n";
	exit 1;
}
binmode($fout);

# channel pointer
print {$fout} "CH"."\x00"x14;
$addr += 16;

# voice
my @ttl;
my @slot = (0,2,1,3);
foreach my $voice (@voice_datas) {
	print {$fout} "AT";
	for(my $y=0; $y<7; $y++) {
	for(my $x=0; $x<4; $x++) {
		my $n = $y * 4 + $slot[$x] + 1;
		print {$fout} pack("C", $voice->[$n]);
		if ($y == 1) {
			$ttl[$slot[$x]]=$voice->[$n];
		}
	}
	}
	print {$fout} pack("C", $voice->[29]);
	print {$fout} "\x00";
	$addr += 32;
}

# title
print {$fout} "D\x14".$info_datas{title}."\x00";
$addr += (2 + length($info_datas{title}) + 1);

print {$fout} "E\x03".$info_datas{credit}."\x00";
$addr += (2 + length($info_datas{credit}) + 1);

# MML parser
my @addr_ch = (0, 0, 0, 0, 0, 0);
my $ch = 0;
foreach my $score_data (@score_datas) {
	last if ($ch >= 6);

	print {$fout} "BS";
	$addr += 2;

	$addr_ch[$ch] = $addr;

	mml_parse_one($ch, $score_data);

	print {$fout} "\x00\x00";
	$addr += 2;

	printf "\nEND: %X\n",$addr;

	$ch++;
}

# write pointer to score datas
seek($fout, 2, 0);
for(my $ch=0; $ch<6; $ch++) {
	my $offset = 0;
	if ($addr_ch[$ch] != 0) {
		$offset = $addr_ch[$ch]-$addr_start-2-$ch*2;
	}
	print {$fout} pack("CC",($offset / 256),($offset & 255));
}

close($fout);

exit 0;

sub mml_parse_one {
	my($ch, $datas)=@_;
	my $atmp = 16;
	my $avol = 12;
	my $aocv = 4;
	my $alen = 4;
	my $sco = 0;
	my $aqva = 8;
	my $aamp = 0;
	my $aems = 1;
	my $aeml = 255;
	my $ajmp = 0;

	my $cmd = 0;

	my $sec = 0;
	foreach my $data (@$datas) {
		my $atlen = 0;
		my $slen = 256;
		for(my $pos=0; $pos < length($data);) {
			my $c = substr($data, $pos, 1);
			$pos++;
			$c = uc($c);

			if ($c ge "A" && $c le "G" || $c eq "R") {
				my $scon = (unpack("C",$c) - 0x42)*2;
				my @alenn = ($alen);
				my $aampn = 0;
				my $aflt = 0;
				my $aerr = 0;
				my $prm = 0;
				my $rprm = 0;

				while(1) {
					my $arg = substr($data, $pos);
					if ($arg =~ /^[+#]/) {
						$aflt = 1;
						$scon = $scon + 1;
						$c = $c."+";
						$pos++;
					} elsif ($arg =~ /^[-]/) {
						$aflt = 1;
						$scon = $scon - 1;
						$c = $c."-";
						$pos++;
					} elsif ($arg =~ /^\./) {
						push(@alenn, $alenn[0] * 2);
						$aampn = 1;
						$c = $c.".";
						$pos++;
					} elsif ($arg =~ /^\&/) {
						$aampn = 1;
						$c = $c."&";
						$pos++;
					} elsif ($arg =~ /^(\d+)/) {
						my $d = $1;
						$alenn[0] = $d;
						$c = $c.$d;
						$pos += length($d);
					} else {
						last;
					}
				}

				if ($scon < 2) {
					$scon = $scon + 14;
				}
				if ($scon > 16) {
					$scon = 16;
				}

				foreach my $alenn (@alenn) {
					# calc_length
					my $prmn = 64/$alenn;
					if ($aampn != 1 && $scon < 16) {
						$prm = int($prmn * $aqva / 4);
						$rprm = int($prmn * 2) - $prm;
					} else {
						$prm = int($prmn * 2);
						$rprm = 0;
					}

					if (($aamp != 0 || $sco == 16) && $sco == $scon) {
						$cmd = 17
					} else {
						$cmd = $scon;
					}

					$sco = $scon;
					$aamp = $aampn;
					$aampn = 0;

					printf "%s SC=%d(%d) LEN=%d PRM=%d \&=%d",$c,$cmd,$sco,$alenn,$prm,$aamp if ($verb);


					if ($cmd == 17 && $prm + $slen <= 256) {
						$slen += $prm;
						$atlen = add_totallen($atlen, $prm);
						seek($fout, -1, 1);
						rep_data8($slen - 1);
					} else {
						put_data(\$cmd, \$prm, \$slen, \$atlen);
					}
					if ($rprm > 0) {
						$cmd = 16;
						$prm = $rprm;
						$rprm = 0;
						printf "= (%d) PRM=%d",$cmd,$prm if ($verb);
						put_data(\$cmd, \$prm, \$slen, \$atlen);
					}
				}

			} elsif ($c eq "<") {
				$slen = 256;
				$aocv--;
				printf "%s AOCV=%d\n",$c,$aocv if ($verb);
				$cmd = 20;
				put_data8($cmd, $aocv);
			} elsif ($c eq ">") {
				$slen = 256;
				$aocv++;
				printf "%s AOCV=%d\n",$c,$aocv if ($verb);
				$cmd = 20;
				put_data8($cmd, $aocv);
			} elsif ($c eq "O") {
				$slen = 256;
				my $arg = get_digit(\$data, \$pos);
				$arg = 4 if ($arg < 1 || $arg > 8);
				$aocv = $arg;
				printf "%s AOCV=%d\n",$c,$aocv if ($verb);
				$cmd = 20;
				put_data8($cmd, $aocv);
			} elsif ($c eq "L") {
				$slen = 256;
				my $arg = get_digit(\$data, \$pos);
				my $prm;
				if ($arg >= 1) {
					$prm = 64 / $arg;
					$alen = $arg;
				} else {
					$prm = 64 / $alen;
				}
				printf "%s ALEN=%d PRN=%.2f\n",$c,$alen,$prm if ($verb);
			} elsif ($c eq "V") {
				$slen = 256;
				my $arg = get_digit(\$data, \$pos);
				$arg = 12 if ($arg < 0 || $arg > 15);
				$avol = $arg;
				if ($ch < 3) {
					# opn
					$cmd = 18;
					printf "%s AVOL=%d\n",$c,$avol if ($verb);
					my @dat = (0, 0, 0, 0);
					for(my $i=0; $i<4; $i++) {
						$dat[$i]=int(127-log($avol+1)/log(16)*(127-$ttl[$i])+0.5);
					}
					put_data32($cmd, \@dat);
				} else {
					# ssg
					$cmd = 19;
					printf "%s AVOL=%d\n",$c,$avol if ($verb);
					put_data8($cmd, $avol);
				}
			} elsif ($c eq "Q") {
				$slen = 256;
				my $arg = get_digit(\$data, \$pos);
				$arg = 8 if ($arg < 1 || $arg > 8);
				$aqva = $arg;
				printf "%s AQVA=%d\n",$c,$aqva if ($verb);
			} elsif ($c eq "S") {
				$slen = 256;
				my $arg = get_digit(\$data, \$pos);
				if ($ch >= 3) {
					$arg = 1 if ($arg < 0 || $arg > 15);
					$cmd = 23;
					$aems = $arg;
					printf "%s AEMS=%d\n",$c,$aems if ($verb);
					put_data8($cmd, $aems);
				}
			} elsif ($c eq "M") {
				$slen = 256;
				my $arg = get_digit(\$data, \$pos);
				if ($ch >= 3) {
					$arg = 0 if ($arg < 0 || $arg > 65535);
					$cmd = 24;
					$aeml = $arg;
					printf "%s AEML=%d\n",$c,$aeml if ($verb);
					put_data16($cmd, $aeml);
				}
			} elsif ($c eq "[") {
				$slen = 256;
				$ajmp = $addr;
				printf "%s ADR=%X\n",$c,$ajmp if ($verb);
			} elsif ($c eq "]") {
				$slen = 256;
				$cmd = 31;
				my $ofs = 0;
				$ofs = $ajmp - $addr - 3 if ($ajmp > 0);
				printf "%s JMP=%X OFS=%d\n",$c,$ajmp,$ofs if ($verb);
				put_data16($cmd, $ofs);
			} elsif ($c eq "T") {
				$slen = 256;
				my $arg = get_digit(\$data, \$pos);
				$arg = 120 if ($arg < 32 || $arg > 255);
				$cmd = 21;
				$atmp=int(32*60/$arg+0.5)+1;
				printf "%s ATMP=%d\n",$c,$atmp if ($verb);
				put_data8($cmd, $atmp);
			} elsif ($c eq "\@") {
				$slen = 256;
				my $arg = get_digit(\$data, \$pos);
				$cmd = 22;
				my $tne = $arg - 1;
				for(my $i=0; $i<4; $i++) {
					$ttl[$slot[$i]] = $voice_datas[$tne][5 + $i];
				}
				if ($verb) {
					printf "%s TNE=%d\n",$c,$tne;
				}
				put_data8($cmd, $tne);
			} elsif ($c eq "Y") {
				$slen = 256;
				my $yaddr = get_digit(\$data, \$pos);
				$yaddr = 0 if ($yaddr < 0 || $yaddr > 255);
				$pos++;
				my $ydat = get_digit(\$data, \$pos);
				$ydat = 0 if ($ydat < 0 || $ydat > 255);
				$cmd = 25;
				printf "%s YADR=%X YDAT=%X\n",$c,$yaddr,$ydat if ($verb);
				put_data16($cmd, $yaddr * 256 + $ydat);
			}

		}
		print "\n" if ($verb);
		printf "**** CH=%d SEC=%d LENGTH=%d\n",$ch+1,$sec+1,$atlen;
		print "\n" if ($verb);
		$sec++;
	}
}

sub put_data
{
	my($rcmd, $rprm, $rslen, $ratlen)=@_;
	while($$rprm >= 0) {
		$$rslen = ($$rprm & 255);
		$$ratlen = add_totallen($$ratlen, $$rslen);
		put_data8($$rcmd, $$rslen - 1);
		$$rprm = $$rprm - 256;
		$$rcmd = 17;
	}
}

sub add_totallen
{
	my($atlen, $dat)=@_;
	$atlen += $dat;
	printf " TLEN=%d\n",$atlen if ($verb);
	return $atlen;
}

sub get_digit
{
	my($rdata, $rpos)=@_;
	my $arg = substr($$rdata, $$rpos);
	my $d = "";
	if ($arg =~ /^(\d+)/) {
		$d = $1;
		$$rpos += length($d);
	}
#	printf "%s -> %s %d\n",$arg,$d,length($d);
	return $d;
}

sub put_data8
{
	my($cmd, $dat)=@_;
	print {$fout} pack("CC",$cmd,$dat);
	printf "  ADR=%X CMD=%X DAT=%X\n",$addr,$cmd,$dat if ($verb);
	$addr += 2;
}

sub put_data16
{
	my($cmd, $dat)=@_;
	$dat += 65536 if ($dat < 0);
	print {$fout} pack("CCC",$cmd,($dat / 256),($dat & 255));
	printf "  ADR=%X CMD=%X DAT=%X\n",$addr,$cmd,$dat if ($verb);
	$addr += 3;
}

sub put_data32
{
	my($cmd, $rdat)=@_;
	print {$fout} pack("C",$cmd);
	printf "  ADR=%X CMD=%X",$addr,$cmd if ($verb);
	for (my $i=0; $i<4; $i++) {
		print {$fout} pack("C",$rdat->[$i]);
		printf " DAT=%X",$rdat->[$i] if ($verb);
	}
	print "\n";
	$addr += 5;
}

sub rep_data8
{
	my($dat)=@_;
	print {$fout} pack("C",$dat);
	printf "  ADR=%X DAT=%X\n",$addr,$dat if ($verb);
}

sub set_datas
{
#
$outfile = "GRA1.bin";
#
%info_datas = (
	'title' => 'GRADIUS STAGE1 BGM - Challenger 1985 - (C)Konami',
	'credit' => 'Original MML was composed by YK-2, and arranged by Sasaji'
);
# ' VOICE DATA x0 x8 x4 xC
@voice_datas = ([
	'Like PSG',		# Comment
	0x02,0x01,0x02,0x01,	# $3x DETUNE & MULTIPLE
	0x1B,0x23,0x2B,0x00,	# $4x TOTAL LEVEL
	0x5F,0x5F,0x1F,0x9F,	# $5x ATTACK RATE & KEY SCALE
	0x01,0x01,0x01,0x01,	# $6x DECAY RATE
	0x01,0x01,0x01,0x01,	# $7x SUSTAIN RATE
	0xF4,0x14,0x14,0x14,	# $8x SUSTAIN LEVEL & RELEASE RATE
	0x00,0x00,0x00,0x00,	# $9x ENVELOP CONTROL
	0x39			# $B0 FEEDBACK & CONTROL(ALGORITUM)
]);
@score_datas = ([
# CH1
	'@1T150V14O5Q8[',
	'L8GRL16G&G&G>C FRE8&R8CR L16G&GR<B-L8B-&B- RL16GA-B->E-FG',
	'C&CR<ARR>C8& C16RFR<AB->CD E-RC<A-R>E-RC FRD<B-R>DFR<',
	'L8GRL16G&G&G>CFRE8&R8CRL16G&GR<B-L8B-&B-RL16GA-B->E-FG',
	'ARC8&CRFRGRARRRFRD&D&DGL8G&G&G&G&G&G',
	'V15L16E-&E-RC&CRL8E-&E-&E-L16CDE-RD&DR<B-&B-RL8G&G&G&G&G',
	'>L16C&C&R<A&AR>CRD&DR<B&BR>DRE-&E-RC&CRE-F&F&FRD&D&DL8G&G&G&G&G&G&G&G&G<V14]',
],[
# CH2
	'@1T150Q8[',
	'V12L16O8C<BAGFGAB>C<BAGFGAB>C<BAGFGAB>C<BAGFGAB>',
	'L16O8C<BAGFGAB>C<BAGFGAB>C<BAGFGAB>CDE-DC<BAB>',
	'V12L16O8C<BAGFGAB>C<BAGFGAB>C<BAGFGAB>C<BAGFGAB>',
	'L16O8C<BAGFGAB>C<BAGFGAB>DC<BAGAB>CDC<BAGB>DG',
	'V15O6L16C&CR<A-&A-R>L8A-&A-&A-&A-&A-L16<B-&B->RD&DRL8E-&E-&E-&E-&E-',
	'L16<AB>CF&F8&FR<B>CDG&G8&GRCDE-A-&A-8&A-RDE-FB-&B-&B->L8C&C&C&C&C<B&B&B&B]',
],[
# CH3
	'@1T150V15Q4[',
	'O4L16C8>CC<C8>CC<C8>CC<C8>CC<E-8>E-E-<E-8>E-E-<E-8>E-E-<E-8>E-E-',
	'O4L16F8>FF<F8>FF<F8>FF<F8>FF<A-8>A-A-<A-8>A-A-<B-8>B-B-<B-8>B-B-',
	'O4L16C8>CC<C8>CC<C8>CC<C8>CC<E-8>E-E-<E-8>E-E-<E-8>E-E-<E-8>E-E-',
	'O4L16F8>FF<F8>FF<F8>FF<F8>FF<G8>GG<G8>GG<G8>GG<G8>GG',
	'L16<<A->A-E-A-<A->A-E-A-<A->A-E-A-<A->A-E-A-<E->E-<B->E-<E->E-<B->E-<E->E-<B->E-<E->E-<B->E-',
	'L16<F>FCF<F>FCF<G>GDG<G>GDG<A->A-E-A-<A->A-E-A-<B->B-FB-<B->B-FB-<G8>GG<G8>GG<G8>GG<G8>GG]',
],[
# CH4 SSG
	'T150V12O4S9[',
	'M8000L8GRL16G&G&G>C FRE8&R8CR L16G&GR<B-L8B-&B- RL16GA-B->E-FG',
	'C&CR<ARR>C8& C16RFR<AB->CD E-RC<A-R>E-RC FRD<B-R>DFR<',
	'M8000L8GRL16G&G&G>CFRE8&R8CRL16G&GR<B-L8B-&B-RL16GA-B->E-FG',
	'ARC8&CRFRGRARRRFRD&D&DGM40000L8G&G&G&G&G&GM10000',
	'>L16E-&E-RC&CRM40000L8E-&E-&E-M10000L16CDE-RD&DR<B-&B-RM40000L8G&G&G&G&G',
	'>L16C&C&R<A&AR>CRD&DR<B&BR>DRE-&E-RC&CRE-F&F&FRD&D&DM50000L8G&G&G&G&G&G&G&G&G<<]',
]);

}
